# start CPU, gather BIOS information, set video mode, load sectors of long mode setup code

#include <lib/memory_layout.h>
#include <lib/mmu.h>


# the BIOS starts executing from here
.global start
start:
        .code16
        cli

        # zero data segment registers
        xorw    %ax, %ax
        movw    %ax, %ds
        movw    %ax, %es
        movw    %ax, %ss

        # enable A20 so physical addresses above 1MB do not wrap
        # TODO: query before enabling?
        movw    $0x2401, %ax
        int     $0x15 

        # read next sectors of long mode setup to 0x7E00 address using BIOS
        leaw    disk_address_packet, %si
        movb    $0x42, %ah
        movb    $0x80, %dl # 1st drive
        int     $0x13

        jmp switch_to_long_mode


.align 4
disk_address_packet:
        .byte   0x10    # size of packet
        .byte   0x0     # unused
        .word   0x7     # number of sectors to read (up until page boundary)
        .word   0x7E00  # offset
        .word   0x0     # segment
        .quad   0x1     # sector number

# MBR
mbr:
        .space (510 - (mbr - start)), 0x0
        .byte 0x55, 0xAA




# switch to 64-bit mode, and jump to C which loads kernel
# reference: AMD64 Manual, Volume 2: System Programming

.set KERNEL_CODE_SEGMENT, 0x8 # second entry

.align 512 # sector boundary
.global switch_to_long_mode
switch_to_long_mode:

        # set VBE video mode
        movw    $0x4F02, %ax
        movw    $0x411B, %bx
        int     $0x10


        # enable Physical Address Extension and Page Global Enabled (to avoid trashing TLB for kernel mappings)
        # TODO: enable global pages in entry.S
        movw    $(CR4_PAE | CR4_PGE), %ax
        movl    %eax, %cr4


        # TODO: we should zero out the memory
        # we map 16 MiB at address 0x200000 and address KERNEL_BASE (2 entries in page table)
        # build Page-Map Level-4 at 0x1000 (free to use; 4096 page aligned)
        movw    $((KERNEL_BASE >> 39) & 0x1FF)*8, %bx
        addw    $0x1000, %bx
        movw    $((0x2000 & 0xF000) | PTE_P | PTE_W), (%bx)
        movw    $(0x2000 >> 16), 2(%bx)
        movw    $0x0, 4(%bx)
        movw    $0x0, 6(%bx)

        # build Page-Directory-Pointer at 0x2000 (4096 page aligned)
        movw    $((KERNEL_BASE >> 30) & 0x1FF)*8, %bx
        addw    $0x2000, %bx
        movw    $((0x3000 & 0xF000) | PTE_P | PTE_W), (%bx)
        movw    $(0x3000 >> 16), 2(%bx)
        movw    $0x0, 4(%bx)
        movw    $0x0, 6(%bx)

        # build Page-Directory at 0x3000 (4096 page aligned)
        # map at 32 MiB in physical memory for kernel (16 pages)
        # TODO: map as much as needed in boot/main.c
        movw    $((KERNEL_BASE >> 21) & 0x1FF)*8, %bx
        addw    $0x3000, %bx
        movw    $0x20, %ax
pd_loop:
        movw    $(PTE_P | PTE_W | PTE_PS), (%bx)
        movw    %ax, 2(%bx)
        movw    $0x0, 4(%bx)
        movw    $0x0, 6(%bx)
        addw    $0x20, %ax
        addw    $8, %bx
        cmp     $(0x3000 + 8*16), %bx
        jl      pd_loop


        # we map also [0, 2 MiB) at address 0x0 (identity map so we do not page fault when enabling paging)
        # PML4
        movw    $0x1000, %bx
        movw    $((0x4000 & 0xF000) | PTE_P | PTE_W), (%bx)
        movw    $(0x4000 >> 16), 2(%bx)
        movw    $0x0, 4(%bx)
        movw    $0x0, 6(%bx)

        # build Page-Directory-Pointer at 0x4000 (4096 page aligned)
        movw    $0x4000, %bx
        movw    $((0x5000 & 0xF000) | PTE_P | PTE_W), (%bx)
        movw    $(0x5000 >> 16), 2(%bx)
        movw    $0x0, 4(%bx)
        movw    $0x0, 6(%bx)

        # build Page-Directory at 0x5000 (4096 page aligned)
        # map at 2 MiB in physical memory
        movw    $0x5000, %bx
        movw    $(PTE_P | PTE_W | PTE_PS), (%bx)
        movw    $0x0, 2(%bx)
        movw    $0x0, 4(%bx)
        movw    $0x0, 6(%bx)


        # load page table
        movw    $0x1000, %ax
        movl    %eax, %cr3


        # update Model Specific Register at address 0xC0000080 with Long Mode Enable
        movl    $0xC0000080, %ecx
        rdmsr
        orl     $EFER_LME, %eax
        wrmsr

        # enable paging and protection (as we are going from real mode to long mode)
        movl    %cr0, %eax
        orl     $(CR0_PG | CR0_WP | CR0_PE), %eax
        movl    %eax, %cr0

        # load Global Descriptor Table
        lgdt    gdt_descriptor

        # long jump with %cs loaded with KERNEL_CODE_SEGMENT
        ljmp    $KERNEL_CODE_SEGMENT, $long_mode


long_mode:
        .code64

        xorq    %rax, %rax
        movq    %rax, %ds
        movq    %rax, %es
        movq    %rax, %fs
        movq    %rax, %gs
        movq    %rax, %ss


        # setup stack pointer
        movq    $start, %rsp
        # jump to C code
        call    boot_main


# hang kernel
spin:
        jmp spin


# global descriptor table (GDT)
.align 8
gdt:
        SEGMENT_NULL
        SEGMENT_KERNEL_CODE

gdt_descriptor:
        .word   0xF // sizeof gdt - 1
        .quad   gdt
