.set KERNEL_CODE_SEGMENT, 0x8 # second entry

# AMD64 still need segments for managing rings, but that's it
#define SEGMENT_NULL \
        .word 0, 0; \
        .byte 0, 0, 0, 0

#define SEGMENT_KERNEL_CODE \
        .word 0, 0; \
        .byte 0, 0x98 /* present bit and required 1 bit */, 0x20 /* long mode */, 0

// Control Register 0
#define CR0_PE  0x1             // Protected Mode Enable
#define CR0_WP  0x10000         // Write Protect (kernel can't write to read-only pages)
#define CR0_PG  0x80000000      // Paging

// Control Register 4
#define CR4_PAE         0x20     // Physical Address Extension

// Page Table Entry
#define PTE_P   0x1ul     // Present
#define PTE_W   0x2ul     // Write
#define PTE_PS  0x80ul    // Page Size

// Extended Feature Enable Register
#define EFER_LME 0x100 // Long Mode Enable



# the BIOS starts executing from here
.global start
start:
        .code16
        cli

        # zero data segment registers
        xorw    %ax, %ax
        movw    %ax, %ds
        movw    %ax, %es
        movw    %ax, %ss

        # enable A20 so physical addresses above 1MB do not wrap
        movw    $0x2401, %ax
        int     $0x15 

        # enable Physical Address Extension
        movw    $CR4_PAE, %ax
        movl    %eax, %cr4


        # map [0, 2 MiB) at address 0x0 (identity map so we do not page fault when enabling paging)
        # PML4
        movw    $0x1000, %bx
        movw    $((0x4000 & 0xF000) | PTE_P | PTE_W), (%bx)
        movw    $(0x4000 >> 16), 2(%bx)
        movw    $0x0, 4(%bx)
        movw    $0x0, 6(%bx)

        # build Page-Directory-Pointer at 0x4000 (4096 page aligned)
        movw    $0x4000, %bx
        movw    $((0x5000 & 0xF000) | PTE_P | PTE_W), (%bx)
        movw    $(0x5000 >> 16), 2(%bx)
        movw    $0x0, 4(%bx)
        movw    $0x0, 6(%bx)

        # build Page-Directory at 0x5000 (4096 page aligned)
        # map at 2 MiB in physical memory
        movw    $0x5000, %bx
        movw    $(PTE_P | PTE_W | PTE_PS), (%bx)
        movw    $0x0, 2(%bx)
        movw    $0x0, 4(%bx)
        movw    $0x0, 6(%bx)


        # load page table
        movw    $0x1000, %ax
        movl    %eax, %cr3


        # update Model Specific Register at address 0xC0000080 with Long Mode Enable
        movl    $0xC0000080, %ecx
        rdmsr
        orl     $EFER_LME, %eax
        wrmsr

        # enable paging and protection (as we are going from real mode to long mode)
        movl    %cr0, %eax
        orl     $(CR0_PG | CR0_WP | CR0_PE), %eax
        movl    %eax, %cr0

        # load Global Descriptor Table
        lgdt    gdt_descriptor

        # long jump with %cs loaded with KERNEL_CODE_SEGMENT
        ljmp    $KERNEL_CODE_SEGMENT, $long_mode


# we are running in long mode (64 bit)
long_mode:
        .code64

        xorq    %rax, %rax
        movq    %rax, %ds
        movq    %rax, %es
        movq    %rax, %fs
        movq    %rax, %gs
        movq    %rax, %ss

        # setup stack pointer
        movq    $start, %rsp

        # print hello world on VGA
        # physical memory address (and linear because of identity mapping) of VGA buffer
        movq    $0xB8000, %rax


        # clear the 80x25 screen
        movq    $0, %rbx
loop:
        movq    $0, (%rax)
        incq    %rax
        cmp     $(0xB8000 + 80*25), %rax
        jne     loop
        

        # reset screen cursor to the start
        movq    $0xB8000, %rax
        # first byte is background/foreground color, second byte is ASCII letter
        movq    $0x7048, 0(%rax) # H
        movq    $0x0765, 2(%rax) # e
        movq    $0x2F6C, 4(%rax) # l
        movq    $0x3E6C, 6(%rax) # l
        movq    $0xFC6F, 6(%rax) # o

# hang kernel
spin:
        jmp spin


# global descriptor table (GDT)
.align 8
gdt:
        SEGMENT_NULL
        SEGMENT_KERNEL_CODE

gdt_descriptor:
        .word   0xF // sizeof gdt - 1
        .quad   gdt


# master boot record (MBR), without this at the end of disk sector it would not boot
mbr:
        .space (510 - (mbr - start)), 0x0
        .byte 0x55, 0xAA
