#ifndef ACPI_H
#define ACPI_H

#include <lib/type.h>

#include "memory.h"


// parsed and simplified ACPI table
struct acpi {
        uint8_t lapics[64]; // enabled ones // TODO: either dynamic or use core_limit
        uint64_t lapic_count;
        uint64_t lapic_pa;
        uint64_t hpet_pa;
};


// is tightly packed by design
struct acpi_rsdp {
        char signature[8];
        uint8_t checksum;
        char oem_id[6];
        uint8_t revision;
        uint32_t rsdt_address; // TODO: use physical pointer type
        uint32_t length;
        uint64_t xsdt_address; // TODO: pa
        uint8_t extended_checksum;
        uint8_t reserved[3];
};


// system description table header
struct acpi_sdth {
        char signature[4];
        uint32_t length;
        uint8_t revision;
        uint8_t checksum;
        char oem_id[6];
        char oem_table_id[8];
        uint32_t oem_revision;
        uint32_t creator_id;
        uint32_t creator_revision;
};


// root system description table
struct acpi_rsdt {
        struct acpi_sdth header;
        uint32_t entry[];
};


// multiple APIC description table
struct acpi_madt {
        struct acpi_sdth header;
        uint32_t local_interrupt_controller_address;
        uint32_t flags;
        struct interrupt_controller_structure { uint8_t type, length; } entry[];
};

// high precision event timer table
struct acpi_hpet {
        struct acpi_sdth header;
        uint32_t event_timer_block_id;
        uint8_t address_space_id;
        uint8_t register_bit_width;
        uint8_t register_bit_offset;
        uint8_t reserved;
        uint32_t address[2];
        uint8_t number;
        uint8_t main_counter_min[2];
        uint8_t page_protection;
};


struct acpi_proc_lapic {
        uint8_t type;
        uint8_t length;
        uint8_t acpi_proc_uid;
        uint8_t apic_id;
        uint32_t flags;
};


int
memcmp(void const *const ptr1, void const *const ptr2, uint64_t const num)
{
        for (uint64_t i = 0; i < num; ++i) {
                int const difference = ((char *)ptr1)[i] - ((char *)ptr2)[i];
                if (difference != 0)
                        return difference;
        }
        return 0;
}


uint8_t
acpi_sum(uint8_t count, uint8_t bytes[count])
{
        uint8_t sum = 0;
        for (uint8_t i = 0; i < count; ++i)
                sum += bytes[i];
        return sum;
}



// TODO: for UEFI system we would get it from EFI_SYSTEM_TABLE
// locate root system descriptor pointer
static inline struct acpi_rsdp *
acpi_find_rsdp(uint64_t const zero_la) // mapped to physical address 0
{
       // search on 16-byte boundaries 
       // TODO: query BIOS for EBDA address
       for (uint64_t pa = 0xE0000; pa < 0xFFFFF; pa += 16) {
                struct acpi_rsdp *const table = (struct acpi_rsdp *)(zero_la + pa);
                if (memcmp(table->signature, "RSD PTR ", 8) == 0 &&
                    acpi_sum(20, (uint8_t *)table) == 0  &&
                    acpi_sum(sizeof *table, (uint8_t *)table) == 0)
                        return table;
       }

       return 0;
}



static inline int
acpi_parse(struct page_table *const pml4, struct frame_pool *const pool, struct heap *const heap, uint64_t const key, struct capability cap, struct acpi *const acpi)
{
        *acpi = (struct acpi){0};

        // TODO: mark as used in frame pool
        uint64_t va_low = KERNEL_BASE + 512*page_size;
        map(pml4, pool, heap, va_low, 0, page_size, PTE_NX, cap);

        struct acpi_rsdp const *const table = acpi_find_rsdp(va_low);
        if (table == 0) {
                unmap(pml4, heap, va_low, page_size);
                return -1;
        }


        // ACPI 1.0
        if (table->revision == 0) {
                // TODO: mark page frame as mapped
                // map the memory address at KERNEL_BASE + ~1 GiB
                uint64_t pa = table->rsdt_address/page_size*page_size;
                uint64_t va = KERNEL_BASE + 513*page_size;

                frame_alloc(pool, pa/page_size, key);
                map(pml4, pool, heap, va, pa, page_size, PTE_NX, cap);

                struct acpi_rsdt *rsdt = va + table->rsdt_address%page_size;
                if (acpi_sum(sizeof rsdt->header + 4*rsdt->header.length, (uint8_t *)rsdt) != 0) {
                        unmap(pml4, heap, va, page_size);
                        unmap(pml4, heap, va_low, page_size);
                        return -1;
                }
                        

                for (uint32_t i = 0; i < (rsdt->header.length - sizeof rsdt->header)/sizeof *rsdt->entry; ++i) {
                        char const *const signature = (char *)(va + rsdt->entry[i]%page_size);

                        struct acpi_sdth const *const header = (struct acpi_sdth *)(va + rsdt->entry[i]%page_size);
                        if (memcmp(header->signature, "APIC", 4) == 0) {
                                struct acpi_madt *madt = header;

                                acpi->lapic_pa = madt->local_interrupt_controller_address;

                                uint32_t offset = sizeof madt->header + sizeof madt->local_interrupt_controller_address + sizeof madt->flags;
                                for (uint32_t j = 0; offset < madt->header.length; ++j) {
                                        uint8_t const type = *((uint8_t *)madt + offset);
                                        uint8_t const length = *((uint8_t *)madt + offset + 1);

                                        enum { processor_local_apic = 0 };
                                        if (type == processor_local_apic) {
                                                struct acpi_proc_lapic const *const lapic = (struct acpi_proc_lapic *)((uint64_t)madt + offset);
                                                if (lapic->flags & 0x1)
                                                        acpi->lapics[acpi->lapic_count++] = lapic->apic_id;
                                        }

                                        offset += length;
                                }

                        } else if (memcmp(header->signature, "HPET", 4) == 0) {
                                struct acpi_hpet *hpet = header;
                                acpi->hpet_pa = (uint64_t)hpet->address[1] << 32 | (uint64_t)hpet->address[0];
                        }
                }
                unmap(pml4, heap, va, page_size);
        }


        unmap(pml4, heap, va_low, page_size);
        return 0;
}


#endif
