// initialize kernel page table

#ifndef INIT_H
#define INIT_H

#include <lib/mmu.h>
#include <lib/std.h>

#include "memory.h"
#include "panic.h"


// physical address of where kernel binary ends
extern char rodata[], data[], bss[], etext[], erodata[], edata[], end[];


// we may not have recursive page table mapping yet so we use simple offset
// does assume initial contiguous mapping
static inline uint64_t
init_va_to_pa(uint64_t const va)
{
        return va - KERNEL_BASE + page_size;
}

static inline uint64_t
init_pa_to_va(uint64_t const pa)
{
        return pa - page_size + KERNEL_BASE;
}



// assumes the heap is physically contiguous (true for the initial portion of heap)
// used only for building initial kernel page table (not in CR3) as it does not invalidate TLBs
// TODO: canonical addresses
// TODO: we could create struct page_table for the bootloader page table and use `map` from `memory.h`
static inline int
init_map(struct page_table *pml4, struct heap *const heap, uint64_t va_start, uint64_t pa, uint64_t size, uint64_t perm)
{
        for (uint64_t va = va_start; va < va_start + size; va += page_size, pa += page_size) {

                uint64_t const idx = (va >> 39) & 0x1FF;
                // allocate page
                if (!(pml4->entries[idx] & PTE_P)) {
                        struct page_table *pdp;
                        heap_alloc(heap, sizeof *pdp, &pdp);
                        *pdp = (struct page_table){0};

                        pml4->entries[idx] = (uint64_t)pdp | PTE_W | PTE_P | PTE_G;
                        pml4->pa_entries[idx] = init_va_to_pa(pdp->pa_entries) | PTE_W | PTE_P | PTE_G;
                }

                struct page_table *pdp = entry_to_address(pml4->entries[idx]);

                // create PDP entry
                {
                        uint64_t const pdp_idx = (va >> 30) & 0x1FF;
                        // allocate page
                        if (!(pdp->entries[pdp_idx] & PTE_P)) {
                                struct page_table *pd;
                                heap_alloc(heap, sizeof *pd, &pd);
                                *pd = (struct page_table){0};

                                pdp->entries[pdp_idx] = (uint64_t)pd | PTE_W | PTE_P | PTE_G;
                                pdp->pa_entries[pdp_idx] = init_va_to_pa(pd->pa_entries) | PTE_W | PTE_P | PTE_G;
                        }

                        struct page_table *pd = entry_to_address(pdp->entries[pdp_idx]);

                        // create PD entry
                        {
                                uint64_t const pd_idx = (va >> 21) & 0x1FF;

                                // allocate page
                                if (!(pd->entries[pd_idx] & PTE_P)) {
                                        pd->entries[pd_idx] = (pa >> 21 << 21) | perm | PTE_PS | PTE_P; // TODO: should present be in permissions?
                                        pd->pa_entries[pd_idx] = (pa >> 21 << 21) | perm | PTE_PS | PTE_P; // TODO: should present be in permissions?
                                } else {
                                        panic("Overwriting page");
                                }
                        }
                }
        }

        return 0;
}


// TODO: error code
// TODO: we assume pml4 address has 2 MiB of memory (page size); should be replaced with 4k allocator
// bootstrap kernel page table
static int
init_pml4(struct page_table *const pml4, struct heap *const heap)
{
        // initialize PML4
        *pml4 = (struct page_table){0};

        // we offset by one page to avoid stepping on low memory (1 MiB)
        // assumes kernel segments are aligned on 2 MiB pages
        // TODO: bit wastefull as we use 8 MiB
        struct mapping const table[] = {
                /* assume start addresses are aligned on page boundary */
                {KERNEL_BASE, page_size, page_size + (etext - KERNEL_BASE), PTE_G}, // text segment
                {rodata, page_size + round_up((etext - KERNEL_BASE), page_size), page_size + (erodata - KERNEL_BASE), PTE_NX | PTE_G}, // read only segment
                {data, page_size + round_up((erodata - KERNEL_BASE), page_size), page_size + (end - KERNEL_BASE) + 5*page_size, PTE_W | PTE_NX | PTE_G}, // data + bss segments; TODO: heap?
        };

        // insert entries
        // permissions are applied only on the PD entries (so we have 2 MiB granularity and not 1 GiB or 512 GiB)
        for (int i = 0; i < sizeof table/sizeof *table; ++i)
                init_map(pml4, heap, table[i].virtual_address, table[i].physical_start, table[i].physical_end - table[i].physical_start, table[i].permissions);

        return 0;
}



static inline int
init_frames(struct frame_pool *pool, uint64_t const kernel_key)
{
        char buffer[64];

        // initial boot memory pool of 16 MiB
        // 2 MiB for storing pool.pages
        pool->slots = (uint8_t *)round_up(end, page_size); // offset to avoid low memory
        SPRINTF(sizeof buffer, buffer, "Pages %x\n", pool->slots)
        uart_write(buffer);

        pool->start = 0;
        pool->end = 4096;
        pool->slot_size = page_size;
        memset(pool->slots, 0, (pool->end - pool->start)*sizeof *pool->slots);

        // flag low memory page
        frame_alloc(pool, 0, kernel_key);

        // flag pages used by kernel
        uint64_t past_end_page = round_up(end - KERNEL_BASE, page_size)/page_size + 1;
        SPRINTF(sizeof buffer, buffer, "Page past kernel %x\n", past_end_page)
        uart_write(buffer);

        // ignore first page (low memory)
        for (uint64_t i = 1; i < past_end_page; ++i) {
                uint64_t idx;
                frame_find(pool, &idx);
                if (frame_alloc(pool, idx, kernel_key) != frame_success)
                        panic("Could not allocate kernel page");
        }

        // mark page structure page as allocated
        uint64_t idx;
        frame_find(pool, &idx);
        if (frame_alloc(pool, idx, kernel_key) != frame_success)
                panic("Could not allocate page structure page");

        // now we have physical page allocator up and running

        // kernel page directory
        // we use whole page so we can address all 512 GiB of kernel space easily
        frame_find(pool, &idx);
        if (frame_alloc(pool, idx, kernel_key) != frame_success)
                panic("Could not allocate frame for kernel frame bitmap");


        // create kernel heap (2 pages, 4k alignment)
        // TODO: put heap of some size into kernel binary .data segment
        // TODO: this is hacking, make it clear where initial kernel heap ends
        for (uint64_t i = 0; i < (end - KERNEL_BASE); i += page_size) {
                frame_find(pool, &idx);
                frame_alloc(pool, idx, kernel_key);
        }
}




static inline void
init_tss(void)
{
}

void
abort_dd(void)
{
        panic("Whatever exception");
}

void
abort_df(void)
{
        panic("Double fault exception");
}
void
abort_gp(uint64_t err)
{
        char buffer[64];
        SPRINTF(sizeof buffer, buffer, "Error %d\n", err)
        uart_write(buffer);
        panic("General protection fault exception");
}
void
abort_pf(void)
{
        char buffer[64];
        SPRINTF(sizeof buffer, buffer, "Linear address %x\n", store_cr2())
        uart_write(buffer);
        panic("Page fault exception");
}


static inline void
init_idt(struct idt *idt)
{
        void timer_interrupt(void);
        void trap_de_asm(void);
        void abort_gp_asm(void);
        void (*traps[64])(void) = {
                [0] = abort_dd,
                [1] = abort_dd,
                [2] = abort_dd,
                [3] = abort_dd,
                [4] = abort_dd,
                [5] = abort_dd,
                [6] = abort_dd,
                [7] = abort_dd,
                [8] = abort_df,
                [9] = abort_dd,
                [10] = abort_dd,
                [11] = abort_dd,
                [12] = abort_dd,
                [13] = abort_gp_asm,
                [14] = abort_pf,
                [32] = timer_interrupt,
        };

        for (int i = 0; i < 64; ++i) {
                idt->gates[i] = (struct gate_descriptor){
                        .offset0 = (uint64_t)traps[i],
                        .selector = 0x8,
                        .ist = 1,
                        .properties = 0x8E,
                        .offset1 = ((uint64_t)traps[i]) >> 16,
                        .offset2 = ((uint64_t)traps[i]) >> 32,
                };
        }


        struct idtr {
                uint16_t limit;
                uint16_t base_address[4];
        };
        struct idtr idtr;
        idtr.limit = sizeof idt->gates - 1;
        idtr.base_address[0] = (uint64_t)idt->gates;
        idtr.base_address[1] = ((uint64_t)idt->gates) >> 16;
        idtr.base_address[2] = ((uint64_t)idt->gates) >> 32;
        idtr.base_address[3] = ((uint64_t)idt->gates) >> 48;

        load_idt((uint64_t)&idtr);
}


#endif
