#ifndef LAPIC_H
#define LAPIC_H

#include "amd64.h"

#define MSR_APIC_BASE 0x1B
#define MSR_APIC_BASE_BSP 0x100ul
#define MSR_APIC_BASE_EXTD 0x400ul // enable x2APIC mode
#define MSR_APIC_BASE_EN 0x800ul

#define MSR_APIC_ID 0x802ul

#define MSR_APIC_ICR 0x830ul // interrupt command register
#define MSR_APIC_ICR_INIT 0x500ul
#define MSR_APIC_ICR_STARTUP 0x600ul

// timer
#define MSR_APIC_LVT_TIMER 0x832ul
#define MSR_APIC_LVT_TIMER_TSC_DEADLINE 0x40000ul
#define MSR_APIC_TSC_DEADLINE 0x6E0ul

#define MSR_APIC_SPURIOUS_INTERRUPT_VECTOR 0x80Ful
#define MSR_APIC_EOI 0x80Bul
#define MSR_APIC_TASK_PRIORITY 0x808ul
#define MSR_APIC_ERROR_STATUS_REGISTER 0x828ul
#define MSR_APIC_LVT_ERROR_REGISTER 0x837ul


static inline void
lapic_init(void)
{
        uint64_t const apic_base = store_msr(MSR_APIC_BASE);

        // TODO: test if BSP is BSP?
        if (!(apic_base & MSR_APIC_BASE_EN))
                panic("APIC is not enabled\n");

        // enable x2APIC
        load_msr(MSR_APIC_BASE, apic_base | MSR_APIC_BASE_EXTD);

        // allow all interrupts with vector >= 32
        load_msr(MSR_APIC_TASK_PRIORITY, 32/16);

        // spurious interrupts (enables APIC)
        load_msr(MSR_APIC_SPURIOUS_INTERRUPT_VECTOR, 0x100ul);
 
        // mask error interrupts
        load_msr(MSR_APIC_LVT_ERROR_REGISTER, 0x10000);
        load_msr(MSR_APIC_ERROR_STATUS_REGISTER, 0);
        load_msr(MSR_APIC_ERROR_STATUS_REGISTER, 0);

        // setup timer
        load_msr(MSR_APIC_LVT_TIMER, MSR_APIC_LVT_TIMER_TSC_DEADLINE | 32);
}


static inline void
lapic_timer(uint64_t const tsc_rate_ms)
{
        load_msr(MSR_APIC_EOI, 0);
        load_msr(MSR_APIC_TSC_DEADLINE, store_tsc() + tsc_rate_ms);
}


static inline uint64_t
lapic_id(void)
{
        return store_msr(MSR_APIC_ID);
}


static inline void
lapic_start_ap(uint64_t const ap_lapic_id, uint64_t const ap_start_address, uint64_t const tsc_rate_ms)
{
        // send initialization message
        load_msr(MSR_APIC_ICR, (ap_lapic_id << 32) /* destination */ | MSR_APIC_ICR_INIT);

        // 10 ms wait
        {
                uint64_t const start_tsc = store_tsc();
                while ((store_tsc() - start_tsc) < 10*tsc_rate_ms);
        }

        // send startup message
        load_msr(MSR_APIC_ICR, (ap_lapic_id << 32) /* destination */ | MSR_APIC_ICR_STARTUP | (ap_start_address >> 12)); // page slot
}


#endif
