#include <lib/amd64.h>
#include <lib/elf.h>
#include <lib/info.h>
#include <lib/memory_layout.h>
#include <lib/type.h>
#include <lib/std.h>

#include <driver/uart.h>

#include "acpi.h"
#include "amd64.h"
#include "container.h"
#include "core.h"
#include "init.h"
#include "lapic.h"
#include "memory.h"
#include "panic.h"
#include "spinlock.h"
#include "syscall.h"




struct frame_pool pool;
struct heap heap;

uint64_t core_count;

struct containers containers;

struct spinlock uart_lock;

struct page_table *kernel_pml4;

static struct capability kernel_cap;
static uint64_t kernel_key;




void
_print_pt(struct page_table *pt, int level, uint64_t va)
{
        va <<= 9;

        char buffer[128];
        for (int i = 0; i < 512; ++i) {
                if (pt->entries[i] == 0 || level == 3)
                        continue;

                for (int j = 0; j < 8*level; ++j)
                        buffer[j] = ' ';
                buffer[8*level] = '\0';
                uart_write(buffer);

                uint64_t canonical = (((va + i) << 21) & (1ul << 47)) ? 0xFFFF000000000000ul | ((va + i) << 21) : (va + i) << 21;
                uint64_t pa = entry_to_address(pt->pa_entries[i]);
                uint64_t const e = pt->pa_entries[i];
                char const perm[] = {
                        (e & PTE_NX) ? '-' : 'X',
                        (e & PTE_G) ? 'G' : '-',
                        (e & PTE_PS) ? 'S' : '-',
                        (e & PTE_U) ? 'U' : '-',
                        (e & PTE_W) ? 'W' : '-',
                        (e & PTE_P) ? 'P' : '-',
                        '\0',
                };

                if (level != 2)
                        SPRINTF(sizeof buffer, buffer, "Entry %d: perm %s\n", i, perm)
                else
                        SPRINTF(sizeof buffer, buffer, "Entry %d: %x -> %x, perm %s\n", i, canonical, pa, perm)
                                uart_write(buffer);


                // can't refer to physical pages as our laddr calculation is wrong for anything outside kernel heap
                if (level == 2)
                        continue;

                // convert physical to virtual address
                _print_pt(entry_to_address(pt->entries[i]), level + 1, va + i);
        }
}

void
print_pt(struct page_table *pt)
{
        _print_pt(pt, 0, 0);
}




void
bootstrap(void)
{
        kernel_key = rdrand();
        capability_init(&kernel_cap, 7, kernel_key);

        uart_init();

        // TODO: BSS
        uart_lock.locked = 0;

        char buffer[64];
        uart_write("Bootstrap\n");

        /* test if first 16 MiB is mapped */
        SPRINTF(sizeof buffer, buffer, "Last mapped value %x\n", *(char *)(KERNEL_BASE + 8*2*1024*1024 - 1))
                uart_write(buffer);


        init_frames(&pool, kernel_key);

        for (int i = 0; i < 32; ++i) {
                SPRINTF(sizeof buffer, buffer, "slot %d, key %d\n", i, pool.slots[i])
                uart_write(buffer);
        }


        // setup heap
        uint64_t past_end_page = round_up(end - KERNEL_BASE, page_size)/page_size + 1;
        heap = (struct heap){
                .current_offset = 0,
                        .start_va = ((past_end_page + 1)*page_size + KERNEL_BASE),
                        .end_va= ((past_end_page + 3)*page_size + KERNEL_BASE),
                        .pool = pool, // in case we need to grow it
        };

        if (heap_alloc(&heap, sizeof *kernel_pml4, &kernel_pml4) != 0)
                panic("Could not allocate memory for PML4");

        if (init_pml4(kernel_pml4, &heap) != 0)
                panic("Could not create kernel page table");

        // since the kernel is contiguously mapped by bootloader
        uint64_t const pml4_pa = (uint64_t)kernel_pml4->pa_entries - KERNEL_BASE + page_size;
        load_cr3(pml4_pa);
        print_pt(kernel_pml4);


        // interrupt descriptor table
        struct idt *idt;
        heap_alloc(&heap, sizeof *idt, &idt);
        init_idt(idt);

        // setup CPU0 (BSP)
        cores[0] = (struct core){
                .rsp_kernel = KERNEL_BASE,
                        .stack_address = KERNEL_BASE - page_size,
                        .stack_size = page_size,
                        .idt = idt,
                        .running = true,
        };
        core_count = 1;

        // save pointer to core in MSR and load to GS
        load_msr(MSR_KERNEL_GS_BASE, &cores[0]);
        swapgs();

        // allocate kernel stack for the core
        uint64_t idx;
        if (frame_find(&pool, &idx) != frame_success || frame_alloc(&pool, idx, kernel_key) != frame_success)
                panic("Failed to allocate frame for BSP kernel stack");

        if (map(kernel_pml4, &pool, &heap, cores[0].stack_address, idx*pool.slot_size, page_size, PTE_NX | PTE_W, kernel_cap) != 0)
                panic("Failed to map BSP kernel stack");
        print_pt(kernel_pml4);

        // TODO: do not jump directly to "main"
        switch_stack(cores[0].rsp_kernel);
}




int
kill(void)
{
        panic("Killed container");
}




void
main(void)
{
        char buffer[64];
        uart_write("Useless OS\n");


        struct core *core = core_address();


        // setup TSS
        // TODO: create separate stack
        struct tss *tss;
        heap_alloc(&heap, sizeof *tss, &tss);
        *tss = (struct tss){
                .ist1[0] = core->stack_address + 4096,
                .ist1[1] = (core->stack_address + 4096) >> 32,
// TODO:                .io_map_base_address = 0xFFFF,
        };



        // setup GDT (4 entries of 8 bytes and one of 16 bytes)
        struct gdt *gdt;
        // TODO: must be 16 byte aligned (it is as of now)
        heap_alloc(&heap, sizeof *gdt, &gdt);
        *gdt = (struct gdt){
                .entry[1] = 0x0020980000000000ul,
                .entry[2] = 0x0000F20000000000ul,
                .entry[3] = 0x0020F80000000000ul,
                // TSS selector
                .entry[4] = (((uint64_t)tss >> 24) << 56) | (((sizeof *tss - 1) & 0xF) << 48) | ((0x89ul) << 40) | (((uint64_t)tss & 0xFFFFFF) << 16) | ((sizeof *tss - 1) & 0xFFFF),
                .entry[5] = (uint64_t)tss >> 32,
        };
        struct gdtr const gdtr = {
                .limit = sizeof *gdt - 1,
                .base_address[0] = (uint64_t)gdt,
                .base_address[1] = (uint64_t)gdt >> 16,
                .base_address[2] = (uint64_t)gdt >> 32,
                .base_address[3] = (uint64_t)gdt >> 48,
        };
        load_gdt((uint64_t)&gdtr);

        load_tr(4 << 3);


        // setup system calls registers
        load_msr(MSR_STAR, 0x0008000800000000ul); // +8 for sysret, +8 for syscall
        load_msr(MSR_LSTAR, syscall_call);
        load_msr(MSR_SFMASK, RFLAGS_INTERRUPT_FLAG); // disable interrupts during syscalls


        uint64_t const pml4_pa = (uint64_t)kernel_pml4 - KERNEL_BASE + page_size;


        // ACPI
        struct acpi acpi;
        if (acpi_parse(kernel_pml4, &pool, &heap, kernel_key, kernel_cap, &acpi) != 0)
                panic("Failed to parse ACPI table");



        // compute TSC rate
        // TODO: assumes same rate for all core
        {
                uint64_t idx = acpi.hpet_pa/pool.slot_size;
                frame_alloc(&pool, idx, kernel_key);

                map(kernel_pml4, &pool, &heap, KERNEL_BASE + 512*page_size, idx*pool.slot_size, page_size, PTE_W, kernel_cap);

                uint64_t *va = (uint64_t *)(KERNEL_BASE + 512*page_size + acpi.hpet_pa%page_size);
                if ((va[0] >> 13) & 0x1 != 1)
                        panic("HPET: 32 bit counter");

                uint64_t const hpet_period_fs = va[0] >> 32;
                SPRINTF(sizeof buffer, buffer, "HPET: period %d ns\n", hpet_period_fs/1000000ul)
                uart_write(buffer);

                // enable HPET
                va[2] |= 0x1;

                // TODO: rdtsc has some overhead so we use rather long period
                uint64_t const start_hpet = va[30];
                uint64_t const start_tsc = store_tsc();
                // wait 1 ms
                while ((va[30] - start_hpet)*hpet_period_fs < 1000000000000ul);
                core->tsc_rate_ms = store_tsc() - start_tsc;

                spinlock_acquire(&uart_lock);
                SPRINTF(sizeof buffer, buffer, "tsc rate %d/ms\n", core->tsc_rate_ms)
                uart_write(buffer);
                spinlock_release(&uart_lock);

                // disable HPET
                va[2] &= ~0x1;

                unmap(kernel_pml4, &heap, KERNEL_BASE + 512*page_size, page_size);
        }



        // start APs
        lapic_init();

        // map low memory (needed for APs)
        uint64_t idx = 0;
        frame_alloc(&pool, idx, kernel_key);
        map(kernel_pml4, &pool, &heap, 0, 0, page_size, PTE_W, kernel_cap);
        print_pt(kernel_pml4);

        for (uint64_t i = 0; i < acpi.lapic_count; ++i) {
                if (acpi.lapics[i] == lapic_id())
                        continue;

                extern char ap_entry_start[], ap_entry_end[];
                uint64_t const pa = 0x5000ul; // scratch space

                memcpy(pa, ap_entry_start, (uint64_t)ap_entry_end - (uint64_t)ap_entry_start);

                // setup core n (AP)
                cores[core_count] = (struct core){
                        .rsp_kernel = KERNEL_BASE - core_count*2*page_size,
                        .stack_address = KERNEL_BASE - core_count*2*page_size - page_size,
                        .stack_size = page_size,
                        .running = false,
                };

                // TODO: should we have all core stacks visible from other cores?
                uint64_t idx;
                frame_find(&pool, &idx);
                frame_alloc(&pool, idx, kernel_key);

                map(kernel_pml4, &pool, &heap, cores[core_count].stack_address, idx*pool.slot_size, page_size, PTE_NX | PTE_W, kernel_cap);

                // TODO: it seems pml4 must live in low memory when loaded
                memcpy(0x3000ul, kernel_pml4->pa_entries, 4096);

                *(uint64_t *)(pa - 8) = 0x3000ul;
                *(uint64_t *)(pa - 16) = cores[core_count].rsp_kernel;
                *(uint64_t *)(pa - 24) = core_count;

                lapic_start_ap(acpi.lapics[i], pa, core->tsc_rate_ms);

                while (!cores[core_count].running);

                core_count += 1;
        }
        unmap(kernel_pml4, &heap, 0, page_size);


        // info area
        uint64_t info_idx;
        {
                frame_find(&pool, &info_idx);
                frame_alloc(&pool, info_idx, kernel_key);
                map(kernel_pml4, &pool, &heap, KERNEL_INFO_ADDR, info_idx*pool.slot_size, page_size, PTE_W | PTE_NX, kernel_cap);
                struct info *info = (struct info *)(KERNEL_INFO_ADDR);
                info->penalty = 0;
                info->container_idx = 0;
                info->core_idx = core - cores;

        }



        // TODO: share with container alloc syscall
        // allocate container array (shared among processors)
        containers.count = 0;
        containers.capacity = 256;
        if (heap_alloc(&heap, containers.capacity*sizeof *containers.data, &containers.data) != 0)
                panic("Failed to allocate container array");


        // construct page table for container
        struct page_table *pml4;
        if (heap_alloc(&heap, sizeof *pml4, &pml4) != 0)
                panic("core pml4 allocation failure");

        // copy PML4
        *pml4 = (struct page_table){0};
        pml4->entries[509] = kernel_pml4->entries[509];
        pml4->pa_entries[509] = kernel_pml4->pa_entries[509];
        pml4->entries[510] = kernel_pml4->entries[510];
        pml4->pa_entries[510] = kernel_pml4->pa_entries[510];
        pml4->entries[511] = kernel_pml4->entries[511];
        pml4->pa_entries[511] = kernel_pml4->pa_entries[511];



        // create container and run demo in it
        containers.data[containers.count] = (struct container){
                .pml4 = pml4,
                .status = status_runnable,
        };
        struct container *const container = &containers.data[containers.count++];


        // setup virtual memory mappings
        // TODO: we should be locking now as other cores are spinning

        // load demo elf embedded in kernel binary
        // TODO: security, as the ELF may be malformed
        extern char _binary_obj_user_demo_elf_start[], _binary_obj_user_demo_elf_size[];
        struct elf_file_header const *const elf = (struct elf_file_header *)_binary_obj_user_demo_elf_start;

        if (elf->ident[0] != '\x7F' || elf->ident[1] != 'E' || elf->ident[2] != 'L' || elf->ident[3] != 'F')
                panic("Not ELF");

        // must be 64-bit
        if (elf->ident[4] != 2 || elf->machine != 0x3E)
                panic("Not Elf-64");

        // TODO: strict aliasing?
        // TODO: set BSS to zero
        // read program segments for each entry in program header table
        struct elf_program_header *ph = _binary_obj_user_demo_elf_start + elf->phoff;
        for (uint16_t i = 0; i < elf->phnum; ++i) {
                // allocate and map memory
                for (uint64_t va = ph[i].vaddr, copied = 0; va < ph[i].memsz + ph[i].vaddr; va += page_size, copied += page_size) {
                        uint64_t idx;
                        frame_find(&pool, &idx);
                        frame_alloc(&pool, idx, kernel_key);
                        // TODO: validate if we are mapping in user space (and have capability)
                        // TODO: simplify (the multiplication trick is even uglier)
                        uint64_t const perm = ((ph[i].flags & ELF_PF_X) ? 0 : PTE_NX)
                                            | ((ph[i].flags & ELF_PF_W) ? PTE_W : 0)
                                            | PTE_U;
                        map(container->pml4, &pool, &heap, va, idx*pool.slot_size, page_size, perm, kernel_cap);

                        // TODO: well specified kernel location
                        // TODO: map more than 1 page at time
                        map(container->pml4, &pool, &heap, KERNEL_BASE + 2*512*page_size, idx*pool.slot_size, page_size, PTE_W | PTE_NX, kernel_cap);
                        // we map the memory both in kernel and in user space as we can't write to user pagesj
                        if (ph[i].memsz > ph[i].filesz) // .bss; TODO: should we zero out all memory when we allocate? (perf?)
                                memset(KERNEL_BASE + 2*512*page_size, 0, page_size);
                        else if (copied < ph[i].filesz) {
                                memcpy(KERNEL_BASE + 2*512*page_size, _binary_obj_user_demo_elf_start + ph[i].offset + va - ph[i].vaddr, (copied + page_size > ph[i].filesz) ? ph[i].filesz%page_size : page_size);
                        }

                        unmap(container->pml4, &heap, KERNEL_BASE + 2*512*page_size, page_size);
                }

        }

        // read symbols for handling exceptions and interrupts
        struct elf_section_header const *const sh = (struct elf_section_header *)(_binary_obj_user_demo_elf_start + elf->shoff);
        bool timer_found = false;
        for (uint16_t i = 0; i < elf->shnum; ++i) {
                if (sh[i].type != elf_sht_symtab)
                        continue;

                // find symbols for exceptions and interrupts
                char const *const string_table = (char *)(_binary_obj_user_demo_elf_start + sh[sh[i].link].offset);
                SPRINTF(sizeof buffer, buffer, "Elf: type %d, offset %x, link %d, entsize %d\n", sh[i].type, sh[i].offset, sh[i].link, sh[i].entsize)
                uart_write(buffer);

                struct elf_symbol const *const symbol_table = (struct elf_symbol *)(_binary_obj_user_demo_elf_start + sh[i].offset);
                for (uint32_t j = 0; j < sh[i].size/sh[i].entsize; ++j) {
                        SPRINTF(sizeof buffer, buffer, "offset %d, name %s\n", symbol_table[j].name, &string_table[symbol_table[j].name])
                        uart_write(buffer);

                        if (strcmp(&string_table[symbol_table[j].name], "timer_interrupt_asm") == 0) {
                                // TODO: do we need to validate anything?
                                container->timer_interrupt = symbol_table[j].value;
                                timer_found = true;
                        }
                }
        }
        if (!timer_found)
                panic("User init should handle timer interrupt");


        // map user stack
        frame_find(&pool, &idx);
        frame_alloc(&pool, idx, kernel_key);
        map(container->pml4, &pool, &heap, USER_TOP - page_size, idx*pool.slot_size, page_size, PTE_NX | PTE_W | PTE_U, kernel_cap);

        // map user info
        map(container->pml4, &pool, &heap, INFO_ADDR, info_idx*pool.slot_size, page_size, PTE_U | PTE_NX, kernel_cap);

        // setup stack and registers as we would be returning from a system call (by setting RIP to system_call_return)
        spinlock_acquire(&uart_lock);
        print_pt(container->pml4);

        SPRINTF(sizeof buffer, buffer, "Elf: entry %x\n", elf->entry)
        uart_write(buffer);

        spinlock_release(&uart_lock);

 
        // switch to user space
        container->regs.rip = elf->entry;
        container->regs.rsp = USER_TOP;

        // allocate 1 quantum for container
        core->quanta.offset = 0;
        core->quanta.end_offset = 1;
        core->quanta.idxs[0] = 0;

        SPRINTF(sizeof buffer, buffer, "cr0 %x, cr4 %x\n", store_cr0(), store_cr4())
        uart_write(buffer);

        lapic_timer(10*core->tsc_rate_ms);
        // TODO: maybe merge syscall_return into syscalls
        syscall_quanta_alloc(core - cores, 0, 1, container->timer_interrupt);
        // TODO: this should not be special case and instead merge with yield (or better associate interrupt handler with quantum)
        core->rip_timer_interrupt = container->timer_interrupt;
        syscall_return(container->regs.rip, container->regs.rsp);
}




void
ap_main(uint64_t const core_idx)
{
        cores[core_idx].running = true;

        // save pointer to core in MSR and load to GS
        load_msr(MSR_KERNEL_GS_BASE, &cores[core_idx]);
        swapgs();

        // setup system calls registers
        load_msr(MSR_STAR, 0x0008000800000000ul); // +8 for sysret, +8 for syscall
        load_msr(MSR_LSTAR, syscall_call);
        load_msr(MSR_SFMASK, RFLAGS_INTERRUPT_FLAG); // disable interrupts during syscalls

        lapic_init();

        char buffer[64];
        spinlock_acquire(&uart_lock);
        SPRINTF(sizeof buffer, buffer, "Hello from CORE %d\n", core_idx)
        uart_write(buffer);
        spinlock_release(&uart_lock);

        for (;;);
}
