// TODO: store process ID in the page? (we could walk through process page table to check)
// TODO: technically uint32_t is enough for page id (as we use 2MiB pages)
// TODO: in general we need to determine physical address backing a virtual address and
//       we can either walk page table or store the physical address with each virtual address,
//       for now we use recursive page table but may move to explicit PA storing (though then we have
//       extra things to keep in sync)

#ifndef MEMORY_H
#define MEMORY_H

#include <lib/capability.h>
#include <lib/mmu.h>
#include <lib/type.h>

#include "amd64.h"
#include "panic.h"
#include "spinlock.h"


static uint64_t const page_size = 0x200000;


// entries and pa_entries attributes must be kept in sync
struct page_table {
        uint64_t entries[512];
        uint64_t pa_entries[512];
};


struct frame_pool {
        uint64_t *slots; // linear memory address
        uint64_t start, end; // page ids
        uint64_t slot_size;
};

// 4096 byte aligned allocations
struct heap {
        uint64_t current_offset; // TODO: use free list or stack
        uint64_t start_va, end_va;
        struct frame_pool pool; // TODO: do we need heap backed by two frame pools? (then the structure would become DAG :/)
        struct spinlock lock;
};


enum frame_result {
        frame_success,
        frame_used,
        frame_unused,
        frame_unavailable,
        frame_key_error,
};

static uint64_t
round_up(uint64_t const x, uint64_t const to)
{
        // TODO: overflow
        // TODO: check if compiler optimize to not and
        return (x/to + 1)*to;
}


// TODO: slow
static inline enum frame_result
frame_find(struct frame_pool const *const pool, uint64_t *const id)
{
        for (uint64_t i = 0; i < pool->end - pool->start; ++i)
                if (pool->slots[i] == 0) {
                        *id = pool->start + i;
                        return frame_success;
                }
        return frame_unavailable;
}


static inline enum frame_result
frame_alloc(struct frame_pool *pool, uint64_t const idx, uint64_t const key)
{
        if (key == 0)
                return frame_key_error;
        if (idx < pool->start || idx >= pool->end)
                return frame_unavailable;
        if (pool->slots[idx - pool->start] != 0)
                return frame_used;

        pool->slots[idx] = key;
        return frame_success;
}


static inline enum frame_result
frame_free(struct frame_pool *pool, uint64_t const idx, uint64_t const key)
{
        if (idx < pool->start || idx >= pool->end)
                return frame_unavailable;
        if (pool->slots[idx - pool->start] != key)
                return frame_key_error;

        pool->slots[idx - pool->start] = 0;
        return frame_success;
}



// TODO: lock outside? heap per core?
static inline int
heap_alloc(struct heap *h, uint64_t const size, uint64_t *va)
{
        spinlock_acquire(&h->lock);
        if (h->current_offset + size > h->end_va) {
                spinlock_release(&h->lock);
                return -1;
        }

        *va = h->start_va + h->current_offset;
        h->current_offset += round_up(size, 4096); // TODO: maybe only 8 byte aligned?

        spinlock_release(&h->lock);
        return 0;

}



static inline struct page_table *
entry_to_address(uint64_t const address)
{
        return (struct page_table *)((address & (1ul << 51)) ? (0xFFFFFF0000000000ul | ((address >> 12 << 12) & ~PTE_NX)) : ((address >> 12 << 12) & ~PTE_NX));
}



static inline int
va_to_pa(struct page_table const *pml4, uint64_t const va, uint64_t *pa)
{
        struct page_table *pdp = entry_to_address(pml4->entries[(va >> 39) & 0x1FFul]);
        struct page_table *pd = entry_to_address(pdp->entries[(va >> 30) & 0x1FFul]);
        uint64_t entry = pd->entries[(va >> 21) & 0x1FFul];

        // TODO: handles correctly only 2 MiB pages
        // TODO: canonical address
        *pa = ((entry >> 21 << 21) & ~PTE_NX) | (va & 0x1FFFFFul);
        
        return 0;
}




struct mapping {
        uint64_t virtual_address;
        uint64_t physical_start;
        uint64_t physical_end;
        uint64_t permissions;
};



static inline int
map(struct page_table *const pml4, struct frame_pool const *const pool, struct heap *const heap, uint64_t va_start, uint64_t pa, uint64_t size, uint64_t perm, struct capability cap)
{
        for (uint64_t va = va_start, i = pa/page_size; va < va_start + size; va += page_size, pa += page_size, ++i) {
                if (i < pool->start || i >= pool->end || !capability_check(cap, pool->slots[i]))
                        return -1;

                uint64_t const idx = (va >> 39) & 0x1FF;
                // allocate page
                if (!(pml4->entries[idx] & PTE_P)) {
                        struct page_table *pdp;
                        heap_alloc(heap, sizeof *pdp, &pdp);
                        *pdp = (struct page_table){0};

                        uint64_t pdp_pa;
                        va_to_pa(pml4, pdp->pa_entries, &pdp_pa);
                        pml4->entries[idx] = (uint64_t)pdp | PTE_W | PTE_P | PTE_G | perm;
                        pml4->pa_entries[idx] = pdp_pa | PTE_W | PTE_P | PTE_G | perm;
                }

                struct page_table *pdp = entry_to_address(pml4->entries[idx]);


                // create PDP entry
                {
                        uint64_t const pdp_idx = (va >> 30) & 0x1FF;
                        // allocate page
                        if (!(pdp->entries[pdp_idx] & PTE_P)) {
                                struct page_table *pd;
                                heap_alloc(heap, sizeof *pd, &pd);
                                *pd = (struct page_table){0};

                                uint64_t pd_pa;
                                va_to_pa(pml4, pd->pa_entries, &pd_pa);
                                pdp->entries[pdp_idx] = (uint64_t)pd |  PTE_P | PTE_G | PTE_W | perm;
                                pdp->pa_entries[pdp_idx] = pd_pa | PTE_P | PTE_G | PTE_W | perm;
                        }


                        struct page_table *pd = entry_to_address(pdp->entries[pdp_idx]);
                        // create PD entry
                        {
                                uint64_t const pd_idx = (va >> 21) & 0x1FF;

                                // allocate page
                                if (!(pd->entries[pd_idx] & PTE_P)) {
                                        pd->entries[pd_idx] = (pa >> 21 << 21) | perm | PTE_PS | PTE_P; // TODO: should present be in permissions?
                                        pd->pa_entries[pd_idx] = (pa >> 21 << 21) | perm | PTE_PS | PTE_P; // TODO: should present be in permissions?
                                } else
                                        panic("Overwriting page");
                        }
                }
        }

        return 0;
}


static inline int
unmap(struct page_table *const pml4, struct heap *const heap, uint64_t va_start, uint64_t size)
{
        for (uint64_t va = va_start; va < va_start + size; va += page_size) {
                // compute indices
                uint64_t const pml4_idx = (va >> 39) & 0x1FF;
                uint64_t const pdp_idx = (va >> 30) & 0x1FF;
                uint64_t const pd_idx = (va >> 21) & 0x1FF;

                // find tables
                if (!(pml4->entries[pml4_idx] & PTE_P))
                        panic("PML4 entry not present");
                struct page_table *pdp = entry_to_address(pml4->entries[pml4_idx]);

                if (!(pdp->entries[pdp_idx] & PTE_P))
                        panic("PDP entry not present");
                struct page_table *pd = entry_to_address(pdp->entries[pdp_idx]);

                if (!(pd->entries[pd_idx] & PTE_P))
                        panic("PD entry not present");

                // unmap
                pd->entries[pd_idx] = 0;
                pd->pa_entries[pd_idx] = 0;

                // TODO: check if whole pdp or pd is all zeroes than heap_free pdp and pd
                uint64_t i;
                for (i = 0; i < 512; ++i)
                        if (pd->entries[i] & PTE_P)
                                break;
                if (i == 512) {
                        // TODO: free page
                        pdp->entries[pdp_idx] = 0;
                        pdp->pa_entries[pdp_idx] = 0;
                        // TODO: reload cr3
                }

                for (i = 0; i < 512; ++i)
                        if (pdp->entries[i] & PTE_P)
                                break;
                if (i == 512) {
                        // TODO: free page
                        pml4->entries[pml4_idx] = 0;
                        pml4->pa_entries[pml4_idx] = 0;
                        // TODO: reload cr3
                }

        }

        // TODO: invalidate only the specific entry
        uint64_t pml4_pa;
        va_to_pa(pml4, pml4->pa_entries, &pml4_pa);

        load_cr3(pml4_pa);

        return 0;
}


#endif
