/*
Syscalls are thin wrappers around UOS kernel functions that mostly checks arguments.
Each syscall returns int64_t and never touches user memory through pointer.

Return 0 means success.
*/

#ifndef SYSCALL_H
#define SYSCALL_H


#include <driver/uart.h>

#include "container.h"
#include "memory.h"
#include "spinlock.h"


extern struct frame_pool pool;
extern struct heap heap;
extern uint64_t core_count;
extern struct containers containers;
extern struct spinlock uart_lock;
extern struct page_table *kernel_pml4;




int64_t
syscall_logc(char const c)
{
        char const tmp[2] = {c, '\0'};
        uart_write(tmp);
        return 0;
}




// TODO: should take capability
int64_t
syscall_container_alloc(uint64_t const container_idx, uint64_t const rip, uint64_t const rsp,
                        uint64_t const de_fault, uint64_t const pf_fault, uint64_t const xf_fault)
{
        spinlock_acquire(&containers.lock);
        if (containers.capacity == containers.count) {
                spinlock_release(&containers.lock);
                return -1;
        }
        uint64_t const idx = containers.count++;
        spinlock_release(&containers.lock);

        // copy kernel mapping of that particular core
        struct page_table *pml4;
        if (heap_alloc(&heap, sizeof *pml4, &pml4) != 0)
                return -1; // TODO: free container

        *pml4 = (struct page_table){0};
        // TODO: use map function
        pml4->entries[508] = containers.data[container_idx].pml4->entries[508];
        pml4->pa_entries[508] = containers.data[container_idx].pml4->pa_entries[508];

        pml4->entries[509] = kernel_pml4->entries[509];
        pml4->pa_entries[509] = kernel_pml4->pa_entries[509];
        pml4->entries[510] = kernel_pml4->entries[510];
        pml4->pa_entries[510] = kernel_pml4->pa_entries[510];
        pml4->entries[511] = kernel_pml4->entries[511];
        pml4->pa_entries[511] = kernel_pml4->pa_entries[511];


        containers.data[idx] = (struct container){
                .regs.rip = rip,
                .regs.rsp = rsp,
                .pml4 = pml4,
                .de_fault = de_fault,
                .pf_fault = pf_fault,
                .xf_fault = xf_fault,
        };

        return idx;
}




int64_t
syscall_container_free(uint64_t const container_idx)
{
        // TODO: check capability
        // TODO: per core containers?
        struct core *core = core_address();

        // free pages memory

        // free frames memory

        // free pml4

        spinlock_acquire(&containers.lock);
        spinlock_release(&containers.lock);

        for (;;);
}




// TODO: handle the possibility of user moving the container to other core
// TODO: having interrupt handler for quanta is confusing as later alloc can change current handler; we should make it so there is handler per quanta
int64_t
syscall_quanta_alloc(uint64_t const core_idx, uint64_t const container_idx, uint64_t const count, uint64_t const timer_interrupt)
{
        // TODO: check if we have capability to modify container with idx
        if (core_idx >= core_count || container_idx >= containers.count)
                return -1;

        containers.data[container_idx].timer_interrupt = timer_interrupt;

        // TODO: allow container migration
        bool const same_core = &cores[core_idx] == core_address();
        if (!same_core)
                return -1;
        //                spinlock_acquire(&cores[core_idx].lock);

        // TODO: lock if our core is different than core_idx
        cores[core_idx].quanta.idxs[cores[core_idx].quanta.end_offset%quanta_capacity] = container_idx;

        #define MIN(a, b) ((a > b) ? b : a)
        uint64_t const maximum = MIN(quanta_capacity - (cores[core_idx].quanta.end_offset - cores[core_idx].quanta.offset), count);
        #undef MIN
        cores[core_idx].quanta.end_offset += maximum;

        //       if (!same_core)
        //              spinlock_release(&cores[core_idx].lock);

        return 0;
}




int64_t
syscall_quanta_yield(uint64_t const core_idx, uint64_t const container_idx, uint64_t const count)
{
        if (core_idx >= core_count || container_idx >= containers.count)
                return -1;

        if (cores[core_idx].quanta.offset + 1 == cores[core_idx].quanta.end_offset)
                panic("Nothing to run");

        // sync
        struct info *info = (struct info *)KERNEL_INFO_ADDR;
        info->penalty = (info->penalty == 0) ? 0 : info->penalty - 1;
        containers.data[container_idx].penalty = info->penalty;

        // debug print quanta
        char buffer[64];
        spinlock_acquire(&uart_lock);
        for (uint64_t i = cores[core_idx].quanta.offset; i < cores[core_idx].quanta.end_offset; ++i) {
                SPRINTF(sizeof buffer, buffer, "quantum %d, core_idx %d, container idx %d\n", i, core_idx, cores[core_idx].quanta.idxs[i%quanta_capacity]);
                uart_write(buffer);
        }
        spinlock_release(&uart_lock);

        uint64_t idx;
        for (uint64_t i = 0; i < count; ++i) {
                idx = cores[core_idx].quanta.idxs[++cores[core_idx].quanta.offset%quanta_capacity];
                // can't yield more than current slice
                if (container_idx != idx)
                        break;
        }

        if (container_idx == idx)
                return 0;


        // CONTEXT SWITCH
        struct registers *const regs = (struct registers *)(cores[core_idx].rsp_kernel - sizeof containers.data[container_idx].regs);
        containers.data[container_idx].regs = *regs;

        *regs = containers.data[idx].regs;

        SPRINTF(sizeof buffer, buffer, "page table from %x to %x\n", containers.data[container_idx].pml4, containers.data[idx].pml4);
        spinlock_acquire(&uart_lock);
        uart_write(buffer);
        spinlock_release(&uart_lock);


        // swap page table
        // TODO: we could cache physical address
        uint64_t pml4_pa;
        va_to_pa(containers.data[idx].pml4, containers.data[idx].pml4->pa_entries, &pml4_pa);
        load_cr3(pml4_pa);

        SPRINTF(sizeof buffer, buffer, "de %x, pf %x, xf %x, idx %d\n", containers.data[idx].de_fault, containers.data[idx].pf_fault, containers.data[idx].xf_fault, idx);
        spinlock_acquire(&uart_lock);
        uart_write(buffer);
        spinlock_release(&uart_lock);


        // switch to container's exception handlers
        cores[core_idx].idt->gates[0] = (struct gate_descriptor){
                .offset0 = containers.data[idx].de_fault,
                .selector = 0x18,
                .ist = 0,
                .properties = 0x8F,
                .offset1 = containers.data[idx].de_fault >> 16,
                .offset2 = containers.data[idx].de_fault >> 32,
        };
        cores[core_idx].idt->gates[14] = (struct gate_descriptor){
                .offset0 = containers.data[idx].pf_fault,
                .selector = 0x18,
                .ist = 0,
                .properties = 0x8F,
                .offset1 = containers.data[idx].pf_fault >> 16,
                .offset2 = containers.data[idx].pf_fault >> 32,
        };
        cores[core_idx].idt->gates[19] = (struct gate_descriptor){
                .offset0 = containers.data[idx].xf_fault,
                .selector = 0x18,
                .ist = 0,
                .properties = 0x8F,
                .offset1 = containers.data[idx].xf_fault >> 16,
                .offset2 = containers.data[idx].xf_fault >> 32,
        };

        // timer interrupt shorthand
        cores[core_idx].rip_timer_interrupt = containers.data[idx].timer_interrupt;

        // update info
        info->penalty = containers.data[idx].penalty;
        info->container_idx = idx;
        info->core_idx = core_idx;

        // restart timer to prevent container yielding right before interrupt and then the next
        // container will immediately be interrupted
        lapic_timer(10*cores[core_idx].tsc_rate_ms);

        // TODO: return value of the other guy
        return 0;
}




// TODO: check capability if container can request frames
int64_t
syscall_frame_alloc(uint64_t const container_idx, uint64_t const idx, uint64_t const key)
{
        if (frame_alloc(&pool, idx, key) != frame_success)
                return -1;
        return 0;
}




// TODO: merge perm and hash into capability struct
int64_t
syscall_map(uint64_t container_idx, uint64_t va_low, uint64_t pa, uint64_t size, uint64_t perm, uint64_t hash)
{
        char buffer[128];
        SPRINTF(sizeof buffer, buffer, "Map, container idx %d, va_low %x, pa %x, size %d\n", container_idx, va_low, pa, size);
        spinlock_acquire(&uart_lock);
        uart_write(buffer);

        print_pt(containers.data[container_idx].pml4);

        int const ret = map(containers.data[container_idx].pml4, &pool, &heap, va_low, pa, size, PTE_U | PTE_W, (struct capability){.perm = perm, .hash = hash});
        print_pt(containers.data[container_idx].pml4);
        spinlock_release(&uart_lock);

        return ret;
}



uint64_t const syscalls[] = {
        syscall_logc,
        syscall_container_alloc,
        syscall_container_free,
        syscall_quanta_alloc,
        syscall_quanta_yield,
        syscall_frame_alloc,
        syscall_map,
};
enum { syscall_count = sizeof syscalls };


#endif
