#ifndef SHA256_H
#define SHA256_H

#include "type.h"


static inline uint32_t
sha256_rotr(uint32_t const x, uint32_t const n)
{
        return (x >> n) | (x << (32 - n));
}


static inline uint32_t
sha256_ch(uint32_t const x, uint32_t const y, uint32_t const z)
{
        return (x & y) ^ ((~x) & z);
}


static inline uint32_t
sha256_maj(uint32_t const x, uint32_t const y, uint32_t const z)
{
        return (x & y) ^ (x & z) ^ (y & z);
}


static inline uint32_t
sha256_sum0(uint32_t const x)
{
        return sha256_rotr(x, 2) ^ sha256_rotr(x, 13) ^ sha256_rotr(x, 22);
}


static inline uint32_t
sha256_sum1(uint32_t const x)
{
        return sha256_rotr(x, 6) ^ sha256_rotr(x, 11) ^ sha256_rotr(x, 25);
}


static inline uint32_t
sha256_sigma0(uint32_t const x)
{
        return sha256_rotr(x, 7) ^ sha256_rotr(x, 18) ^ (x >> 3);
}


static inline uint32_t
sha256_sigma1(uint32_t const x)
{
        return sha256_rotr(x, 17) ^ sha256_rotr(x, 19) ^ (x >> 10);
}




// big-endian, only messages <= 14 uint32_t
static inline bool
sha256(uint64_t const count, uint32_t const message[count], uint32_t result[8])
{
        if (count > 14)
                return false;


        // constants; section 4.2.2
        static uint32_t const constants[64] = {
                0x428A2F98, 0x71374491, 0xB5C0FBCF, 0xE9B5DBA5, 0x3956C25B, 0x59F111F1, 0x923F82A4, 0xAB1C5ED5,
                0xD807AA98, 0x12835B01, 0x243185BE, 0x550C7DC3, 0x72BE5D74, 0x80DEB1FE, 0x9BDC06A7, 0xC19BF174,
                0xE49B69C1, 0xEFBE4786, 0x0FC19DC6, 0x240CA1CC, 0x2DE92C6F, 0x4A7484AA, 0x5CB0A9DC, 0x76F988DA,
                0x983E5152, 0xA831C66D, 0xB00327C8, 0xBF597FC7, 0xC6E00BF3, 0xD5A79147, 0x06CA6351, 0x14292967,
                0x27B70A85, 0x2E1B2138, 0x4D2C6DFC, 0x53380D13, 0x650A7354, 0x766A0ABB, 0x81C2C92E, 0x92722C85,
                0xA2BFE8A1, 0xA81A664B, 0xC24B8B70, 0xC76C51A3, 0xD192E819, 0xD6990624, 0xF40E3585, 0x106AA070,
                0x19A4C116, 0x1E376C08, 0x2748774C, 0x34B0BCB5, 0x391C0CB3, 0x4ED8AA4A, 0x5B9CCA4F, 0x682E6FF3,
                0x748F82EE, 0x78A5636F, 0x84C87814, 0x8CC70208, 0x90BEFFFA, 0xA4506CEB, 0xBEF9A3F7, 0xC67178F2,
        };

        uint32_t message_block[16] = {0};
        for (uint32_t i = 0; i < count; ++i)
                message_block[i] = message[i];


        // pad message; section 5.1.1
        if (count < 14)
                message_block[count] = 0x80000000;

        message_block[15] = count*32; // TODO: we assume only messages smaller than 64 bytes

        
        // initial hash value; section 5.3.2
        uint32_t hash[8] = {0x6A09E667, 0xBB67AE85, 0x3C6EF372, 0xA54FF53A, 0x510E527F, 0x9B05688C, 0x1F83D9AB, 0x5BE0CD19};


        // hash computation; section 6.2.2
        uint32_t w[64];
        for (int t = 0; t < 16; ++t)
                w[t] = message_block[t];
        for (int t = 16; t < 64; ++t)
                w[t] = sha256_sigma1(w[t - 2]) + w[t - 7] + sha256_sigma0(w[t - 15]) + w[t - 16];

        uint32_t a = hash[0];
        uint32_t b = hash[1];
        uint32_t c = hash[2];
        uint32_t d = hash[3];
        uint32_t e = hash[4];
        uint32_t f = hash[5];
        uint32_t g = hash[6];
        uint32_t h = hash[7];


        for (uint32_t t = 0; t < 64; ++t) {
                uint32_t const t1 = h + sha256_sum1(e) + sha256_ch(e, f, g) + constants[t] + w[t];
                uint32_t const t2 = sha256_sum0(a) + sha256_maj(a, b, c);

                h = g;
                g = f;
                f = e;
                e = d + t1;
                d = c;
                c = b;
                b = a;
                a = t1 + t2;
        }

        hash[0] += a;
        hash[1] += b;
        hash[2] += c;
        hash[3] += d;
        hash[4] += e;
        hash[5] += f;
        hash[6] += g;
        hash[7] += h;

        for (int i = 0; i < 8; ++i)
                result[i] = hash[i];

        return true;
}




#endif
