#include <lib/amd64.h>
#include <lib/capability.h>
#include <lib/info.h>
#include <lib/log.h>
#include <lib/memory_layout.h>
#include <lib/std.h>
#include <lib/type.h>


static uint64_t const page_size = 0x200000ul;


static inline void
draw_rect(uint8_t fb[], uint64_t xres, uint8_t color[3], uint64_t x, uint64_t y, uint64_t width, uint64_t height)
{
        for (uint64_t j = y; j < y + height; ++j)
                for (uint64_t i = x; i < x + width; ++i)
                        for (int p = 0; p < 3; ++p)
                                fb[3*(xres*j + i) + p] = color[2 - p];
}


static inline void
panic(char const message[static 1])
{
        log(message);
        log("\n");
        for (;;);
}


void
de_fault(uint64_t ss, uint64_t rsp, uint64_t rflags, uint64_t cs, uint64_t rip)
{

        char buffer[256];
        SPRINTF(sizeof buffer, buffer, "ss %x\nrsp %x\nrflags %x\ncs %x\nrip %x\n", ss, rsp, rflags, cs, rip)
        log(buffer);

        panic("divide by zero");
}


void
xf_fault(uint64_t ss, uint64_t rsp, uint64_t rflags, uint64_t cs, uint64_t rip)
{

        char buffer[256];
        SPRINTF(sizeof buffer, buffer, "#XF: ss %x\nrsp %x\nrflags %x\ncs %x\nrip %x\n", ss, rsp, rflags, cs, rip)
        log(buffer);

        panic("#XF: SIMD");
}


void
pf_fault(uint64_t ss, uint64_t rsp, uint64_t rflags, uint64_t cs, uint64_t rip, uint64_t error_code)
{

        char buffer[256];
        SPRINTF(sizeof buffer, buffer, "#PF: ss %x\nrsp %x\nrflags %x\ncs %x\nrip %x\nerror code %d\n", ss, rsp, rflags, cs, rip, error_code)
        log(buffer);

        panic("#PF: Page Fault");
}


void
timer_interrupt(void)
{
        struct info const *const info = (struct info *)INFO_ADDR;
        if (info->penalty != 1) {
                char buffer[64];
                SPRINTF(sizeof buffer, buffer, "Penalty not 1, it is %d\n", info->penalty)
                panic(buffer);
        }

        void timer_interrupt_asm(void);
        syscall_quanta_alloc(info->core_idx, info->container_idx, 1, timer_interrupt_asm);
        syscall_quanta_yield(info->core_idx, info->container_idx, 1);
}


int
main(void)
{
        struct info const *const info = (struct info *)INFO_ADDR;

        uint64_t const local_key = 10;
        struct capability shared_cap;
        capability_init(&shared_cap, 7, local_key);

        char buffer[64];
        SPRINTF(sizeof buffer, buffer, "HELLO, core %d, container %d, %d\n\n\n", info->core_idx, info->container_idx, rdrand())//*(uint64_t *)0x1000000)
        log(buffer);


        if (info->container_idx != 0) {
                uint64_t const key = rdrand();
                struct capability cap;
                capability_init(&cap, 7, key);


                // TESTS
                uint64_t frame_idx = 64;
                if (syscall_frame_alloc(info->container_idx, frame_idx, key) != -1)
                        panic("Allocated already allocated frame");
                if (syscall_map(info->container_idx, 0x200000, frame_idx*page_size, page_size, cap.perm, cap.hash) != -1)
                        panic("Mapped frame without having capability");
                if (syscall_map(info->container_idx, 66*page_size, 66*page_size, page_size, shared_cap.perm, shared_cap.hash) == -1)
                        panic("Failed to map shared memory");
                uint64_t *shared_mem = (uint64_t *)(66*page_size);

                void gen_ud_fault(void);
               // gen_ud_fault();
                void gen_int(void);
               // gen_int();

                //if (1/0)
                        *shared_mem = 10;
               // *shared_mem = *(uint64_t *)0x1000000ul;


                uint64_t const x_resolution = 1280;
                uint64_t const y_resolution = 1024;

                uint64_t framebuffer_pa = 0xFD000000ul;
                uint8_t *framebuffer = (uint8_t *)(10*page_size);

                // allocate frames
                for (uint64_t pa = framebuffer_pa; pa < framebuffer_pa + x_resolution*y_resolution*3; pa += page_size)
                        syscall_frame_alloc(info->container_idx, pa/page_size, key);

                if (syscall_map(info->container_idx, framebuffer, framebuffer_pa, x_resolution*y_resolution*3, cap.perm, cap.hash) != 0)
                        log("Framebuffer map failed\n");

                log("Mapped framebuffer\n");


                // simple game of life
                enum { rows = 64*1280/1024, cols = 64, };
                uint8_t grid[rows][cols] = {0,};
                uint8_t tmp[rows][cols] = {0,};

                // blinker
                grid[1][1] = 2;
                grid[1][2] = 2;
                grid[1][3] = 2;

                // from Blow's talk
                uint64_t const roff = rows/2;
                uint64_t const coff = cols - 10;
                grid[roff - 5][coff - 5] = 2;
                grid[roff - 5][coff - 6] = 2;
                grid[roff - 5][coff - 7] = 2;
                grid[roff - 5][coff - 10] = 2;
                grid[roff - 5][coff - 11] = 2;
                grid[roff - 6][coff - 8] = 2;
                grid[roff - 7][coff - 10] = 2;


                uint64_t time = store_tsc();
                for (;;) {
                        *shared_mem = rdrand();

                        // apply rules
                        for (int64_t i = 0; i < rows; ++i)
                        for (int64_t j = 0; j < cols; ++j) {
                                int alive = 0;
                                for (int64_t ni = -1; ni <= 1; ++ni)
                                for (int64_t nj = -1; nj <= 1; ++nj) {
                                        if (i + ni >= 0 && i + ni < rows && j + nj >= 0 && j + nj < cols)
                                                alive += grid[i + ni][j + nj] == 2;
                                };
                                tmp[i][j] = alive - (grid[i][j] == 2);

                        }
                        for (int64_t i = 0; i < rows; ++i)
                        for (int64_t j = 0; j < cols; ++j) {
                                if (grid[i][j] == 2 && (tmp[i][j] < 2 || tmp[i][j] > 3))
                                        grid[i][j] = 1;
                                else if (grid[i][j] != 2 && tmp[i][j] == 3)
                                        grid[i][j] = 2;
                        }


                        // draw grid
                        uint8_t const foreground[3] = {255, 128, 64};
                        uint8_t const background[3] = {250, 250, 250};
                        uint8_t const visited[3] = {180, 255, 128};
                        for (uint64_t i = 0, y = 0; i < rows; ++i, y += y_resolution/rows)
                        for (uint64_t j = 0, x = 0; j < cols; ++j, x += x_resolution/cols) {
                                if (grid[i][j] == 2)
                                        draw_rect(framebuffer, x_resolution, foreground, x, y, x_resolution/cols, y_resolution/rows);
                                else if (grid[i][j] == 1)
                                        draw_rect(framebuffer, x_resolution, visited, x, y, x_resolution/cols, y_resolution/rows);
                                else
                                        draw_rect(framebuffer, x_resolution, background, x, y, x_resolution/cols, y_resolution/rows);
                        }

                        // TODO: sleep for some measurable time
                //        while (store_tsc() - time < 100000000ul);
                        time = store_tsc();
                }
        }

        void fault_de(void);
        void fault_pf(void);
        void fault_xf(void);
        int64_t const idx = syscall_container_alloc(info->container_idx, page_size, 2*page_size, fault_de, fault_pf, fault_xf);
        SPRINTF(sizeof buffer, buffer, "Allocated container %d\n", idx)
        log(buffer);

        uint64_t const key = rdrand();
        struct capability cap;
        capability_init(&cap, 7, key);

        // text
        uint64_t frame_idx = 64;
        syscall_frame_alloc(idx, frame_idx, key);
        syscall_map(idx, 0x200000, frame_idx*page_size, page_size, cap.perm, cap.hash);
        syscall_map(info->container_idx, 64*page_size, frame_idx*page_size, page_size, cap.perm, cap.hash);
        for (uint64_t i = 0; i < page_size; ++i)
                ((uint8_t *)(64*page_size))[i] = ((uint8_t *)page_size)[i];

        // rodata
        frame_idx = 65;
        syscall_frame_alloc(idx, frame_idx, key);
        syscall_map(idx, 2*page_size, frame_idx*page_size, page_size, cap.perm, cap.hash);
        syscall_map(info->container_idx, 65*page_size, frame_idx*page_size, page_size, cap.perm, cap.hash);
        for (uint64_t i = 0; i < page_size; ++i)
                ((uint8_t *)(65*page_size))[i] = ((uint8_t *)(2*page_size))[i];


        // data, bss
        /*
        frame_idx = 67;
        syscall_frame_alloc(idx, frame_idx, key);
        syscall_map(idx, 3*page_size, frame_idx*page_size, page_size, cap.perm, cap.hash);
        syscall_map(info->container_idx, 67*page_size, frame_idx*page_size, page_size, cap.perm, cap.hash);
        for (uint64_t i = 0; i < page_size; ++i)
                ((uint8_t *)(67*page_size))[i] = ((uint8_t *)(3*page_size))[i];
        */

        // shared memory
        frame_idx = 66;
        syscall_frame_alloc(idx, frame_idx, local_key);
        syscall_map(info->container_idx, 66*page_size, frame_idx*page_size, page_size, shared_cap.perm, shared_cap.hash);
        uint64_t *shared_mem = (uint64_t *)(66*page_size);


        // TODO: should serve also for migration?
        void timer_interrupt_asm(void);
        syscall_quanta_alloc(info->core_idx, idx, 1, timer_interrupt_asm);

        for (;;) {
                SPRINTF(sizeof buffer, buffer, "Bootstrap container, %d\n", *shared_mem)
                log(buffer);
        }
}
